__all__ = ['camxfiles_paths', 'net_balance_paths', 'geoschemfiles_paths',
           'icarttfiles_paths', 'all_paths', 'self_described_paths',
           'cmaqfiles_paths']

from os.path import join, abspath

cmaqfiles_paths = dict(
    icon_profile='cmaqfiles/profiles/test.icon_profile',
    bcon_profile='cmaqfiles/profiles/test.bcon_profile',
)

for key, val in cmaqfiles_paths.items():
    cmaqfiles_paths[key] = abspath(join(*__path__ + val.split('/')))

camxfiles_paths = dict(
    wind='camxfiles/wind/test.wind',
    landuse='camxfiles/landuse/test.landuse',
    temperature='camxfiles/temperature/test.temperature',
    cloud_rain='camxfiles/cloud_rain/test.cloud_rain',
    vertical_diffusivity=('camxfiles/vertical_diffusivity/' +
                          'test.vertical_diffusivity'),
    humidity='camxfiles/humidity/test.humidity',
    height_pressure='camxfiles/height_pressure/test.height_pressure',
    uamiv='camxfiles/uamiv/test.uamiv',
    point_source='camxfiles/point_source/test.point_source',
    lateral_boundary='camxfiles/lateral_boundary/test.lateral_boundary')

for key, val in camxfiles_paths.items():
    camxfiles_paths[key] = abspath(join(*__path__ + val.split('/')))

net_balance_paths = dict(
    mrg_file='net_balance/test.mrg_file',
    net_file='net_balance/test.net_file',
)

for key, val in net_balance_paths.items():
    net_balance_paths[key] = join(*__path__ + val.split('/'))

geoschemfiles_paths = dict(bpch='geoschemfiles/test.bpch')

for key, val in geoschemfiles_paths.items():
    geoschemfiles_paths[key] = join(*__path__ + val.split('/'))

icarttfiles_paths = dict(ffi1001=join(
    *__path__ + ['icarttfiles', 'test.ffi1001']))

ceilometerfiles_paths = dict(
    vaisala=join(
        *__path__ +
        ['ceilometerfiles', 'VAISALA_CEILOMETER_1_LEVEL2_20.his']
    )
)

all_paths = dict()
all_paths.update(camxfiles_paths, **geoschemfiles_paths)
all_paths.update(icarttfiles_paths, **net_balance_paths)

self_described_paths = dict()
for k in ['uamiv', 'point_source', 'lateral_boundary', 'humidity',
          'vertical_diffusivity']:
    self_described_paths[k] = camxfiles_paths[k]

self_described_paths['bpch'] = geoschemfiles_paths['bpch']
self_described_paths['bpch2'] = geoschemfiles_paths['bpch']
self_described_paths['ffi1001'] = icarttfiles_paths['ffi1001']



"""
.. _cloud_rain
:mod:`cloud_rain` -- Cloud/Rain File Interfaces
===============================================

.. module:: cloud_rain
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` file interfaces for CAMx cloud/rain
              files.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmap', 'Transforms']

from . import Memmap
from . import Transforms

if __name__ == '__main__':
    from PseudoNetCDF.camxfiles.cloud_rain.Memmap import cloud_rain
    from PseudoNetCDF.pncdump import pncdump_parser, \
        dump_from_cmd_line
    parser = pncdump_parser()
    (file_path, options, extra_args_dict) = parser.parse_args()

    dump_from_cmd_line(file_path, options, cloud_rain)



__all__ = ['ncf2vertical_diffusivity']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx vertical diffusivity  writer
=================================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` writer for CAMx vertical
              diffusivity files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

from PseudoNetCDF.camxfiles.one3d.Write import ncf2one3d
from PseudoNetCDF._getwriter import registerwriter

ncf2vertical_diffusivity = ncf2one3d
registerwriter('camxfiles.vertical_diffusivity', ncf2vertical_diffusivity)
registerwriter('vertical_diffusivity', ncf2vertical_diffusivity)



__all__ = ['one3d']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- one3d Memmap interface
============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx generic
              one 3d variable files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array, where, memmap, newaxis

# This Package modules
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class one3d(PseudoNetCDFFile):
    """
    one3d provides a PseudoNetCDF interface for CAMx
    one3d files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> one3d_path = 'camx_one3d.bin'
        >>> rows,cols = 65,83
        >>> one3dfile = one3d(one3d_path,rows,cols)
        >>> one3dfile.variables.keys()
        ['TFLAG', 'UNKNOWN']
        >>> tflag = one3dfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v = one3dfile.variables['UNKNOWN']
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> one3dfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = "fi"
    data_fmt = "f"
    var_name = "UNKNOWN"
    units = "UNKNOWN"

    @classmethod
    def isMine(cls, path):
        try:
            f = one3d(path)
        except Exception:
            return False
        return f._checkfile()

    def _checkfile(self):
        if self.__memmap.size != (self.__records * self.__record_items):
            return False
        pad0 = self.__memmap.reshape(
            self.__records, self.__record_items).view('i')[:, 0]
        pad1 = self.__memmap.reshape(
            self.__records, self.__record_items).view('i')[:, -1]
        return (pad0 == pad1).all()

    def __init__(self, rf, rows=None, cols=None):
        """
        Initialization included reading the header and learning
        about the format.

        rows - number of rows in domain (defaults: 1)
        cols - number of columns in domain (defaults: size)

        """

        self.rffile = rf

        self.__memmap = memmap(self.rffile, '>f', 'r', offset=0)
        if rows is None and cols is None:
            rows = 1
            cols = self.__memmap[[-1]].view('>i')[0] // 4 - 2
        self.__record_items = rows * cols + 4

        self.__records = self.__memmap.shape[0] // self.__record_items

        time_date = array(self.__memmap.reshape(
            self.__records, self.__record_items)[:, 1:3])

        lays = where(time_date != time_date[newaxis, 0])[0][0]

        new_hour = slice(0, None, lays)

        dates = time_date[:, 1].view('>i')
        times = time_date[:, 0]

        self.__tflag = array(
            [dates[new_hour], times[new_hour]], dtype='>f').swapaxes(0, 1)
        time_steps = self.__records / lays

        self.createDimension('VAR', 1)
        self.createDimension('TSTEP', time_steps)
        self.createDimension('COL', cols)
        self.createDimension('ROW', rows)
        self.createDimension('LAY', lays)
        self.createDimension('DATE-TIME', 2)

        self.FTYPE = 1
        self.NVARS = 1
        self.NCOLS = cols
        self.NROWS = rows
        self.NLAYS = lays
        self.NTHIK = 1

        self.variables = PseudoNetCDFVariables(
            self.__variables, [self.var_name])
        v = self.variables['TFLAG'] = ConvertCAMxTime(
            self.__tflag[:, 0], self.__tflag[:, 1], 1)
        self.SDATE, self.STIME = v[0, 0, :]

    def __decorator(self, name, pncfv):
        props = dict(units=self.units, var_desc=self.var_name.ljust(16),
                     long_name=self.var_name.ljust(16))
        for k, v in props.items():
            setattr(pncfv, k, v)
        return pncfv

    def __variables(self, k):
        pncvar = PseudoNetCDFVariable
        tsteps = len(self.dimensions['TSTEP'])
        lays = len(self.dimensions['LAY'])
        rows = len(self.dimensions['ROW'])
        cols = len(self.dimensions['COL'])
        tmpvals = self.__memmap\
            .reshape(self.__records, self.__record_items)[:, 3:-1]\
            .reshape(tsteps, lays, rows, cols)
        return self.__decorator(k, pncvar(self, k, 'f',
                                          ('TSTEP', 'LAY', 'ROW', 'COL'),
                                          values=tmpvals))


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testKV(self):
        import PseudoNetCDF.testcase
        inpath = PseudoNetCDF.testcase.camxfiles_paths['vertical_diffusivity']
        vdfile = one3d(inpath, 4, 5)
        vdfile.variables['TFLAG']
        checkv = array([1.00000000e+00, 4.76359320e+00, 1.92715893e+01,
                        1.52158489e+01, 7.20601225e+00, 1.84097159e+00,
                        2.63084507e+01, 1.27621298e+01, 1.06348248e+01,
                        2.22587357e+01, 1.00000000e+00, 1.69009724e+01,
                        1.00000000e+00, 2.23075104e+01, 1.27485418e+01,
                        1.45508013e+01, 1.45637455e+01, 2.95294094e+01,
                        3.02676849e+01, 2.84974957e+01, 1.00000000e+00,
                        6.93706131e+00, 3.07418957e+01, 3.41300621e+01,
                        1.66266994e+01, 1.00000000e+00, 2.53920174e+01,
                        1.92539787e+01, 2.32906532e+01, 5.96702042e+01,
                        1.00000000e+00, 2.24458847e+01, 1.00000000e+00,
                        5.45038452e+01, 3.45825729e+01, 3.43578224e+01,
                        3.76071548e+01, 6.76799850e+01, 7.33648529e+01,
                        7.35801239e+01, 1.00000000e+00, 6.38178444e+00,
                        4.39327278e+01, 5.00754166e+01, 2.44474106e+01,
                        1.00000000e+00, 3.51700935e+01, 2.71137428e+01,
                        3.40312347e+01, 8.85775909e+01, 1.00000000e+00,
                        3.13994522e+01, 1.00000000e+00, 8.07169266e+01,
                        5.12892876e+01, 5.05734329e+01, 5.56603966e+01,
                        1.00394188e+02, 1.08980370e+02, 1.09251083e+02,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.64916098e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.22205174e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.05408611e+01,
                        1.26687222e+01, 1.21386652e+01, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.93040049e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.39350021e+00, 1.02697349e+00, 1.00000000e+00,
                        1.00000000e+00, 1.82250175e+01, 2.90407104e+01,
                        2.83827496e+01, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.00000000e+00,
                        1.00000000e+00, 2.02609706e+00, 1.00000000e+00,
                        1.00000000e+00, 1.00000000e+00, 1.44422662e+00,
                        1.02998519e+00, 1.00000000e+00, 1.00000000e+00,
                        2.60322971e+01, 4.26534195e+01, 4.17046585e+01],
                       dtype='f').reshape(2, 3, 4, 5)
        self.assertTrue((vdfile.variables['UNKNOWN'] == checkv).all())

    def testNCF2KV(self):
        import PseudoNetCDF.testcase
        from PseudoNetCDF.pncgen import pncgen
        import os
        inpath = PseudoNetCDF.testcase.camxfiles_paths['vertical_diffusivity']
        outpath = inpath + '.check'
        infile = one3d(inpath, 4, 5)
        pncgen(infile, outpath, format='camxfiles.vertical_diffusivity')
        orig = open(inpath, 'rb').read()
        new = open(outpath, 'rb').read()
        assert(orig == new)
        os.remove(outpath)


if __name__ == '__main__':
    unittest.main()



__doc__ = """
.. _cmaqfiles
:mod:`cmaqfile` -- CMAQ File Interfaces
===================================

.. module:: CMAQ
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF`  random access read
              based file interfaces for all CMAQ related files.  See
              PseudoNetCDF.sci_var.PseudoNetCDFFile for interface details.
              Each camx file level module (i.e. uamiv, cloud_rain, etc) can
              be called as a script for ncdump style output.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['boxmodel', 'pa', 'box_model_mrg', 'box_model_conc', 'jtable',
           'bcon_profile', 'icon_profile', 'ioapi', 'ioapi_base', 'griddesc',
           'cmaqomidat']

from . import boxmodel
from . import pa
from ._jtable import jtable
from .boxmodel import box_model_conc, box_model_mrg
from .profile import bcon_profile, icon_profile
from ._ioapi import ioapi, ioapi_base
from ._griddesc import griddesc
from ._cmaqomidat import cmaqomidat



__all__ = ['ncf2one3d']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx one3d writer
============================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` writer for CAMx generic
              1 3D variable files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

from numpy import array
from PseudoNetCDF._getwriter import registerwriter


def ncf2one3d(ncffile, outpath, key=None, tflag='TFLAG'):
    outfile = open(outpath, 'wb')
    keys = ncffile.variables.keys()
    if key is None:
        key, = [k for k in keys if k != 'TFLAG']
    for (d, t), v3d in zip(ncffile.variables[tflag][:, 0, :],
                           ncffile.variables[key]):
        t = array(t.astype('>f') / 100, ndmin=1).astype('>f')
        d = array(d, ndmin=1).astype('>i')
        d = (d % (d // 100000 * 100000)).astype('>i')
        for v2d in v3d:
            v2d = v2d.astype('>f')
            buf = array((v2d.size + 2) * 4, ndmin=1).astype('>i').tostring()
            outfile.write(buf + t.tostring() + d.tostring() +
                          v2d.tostring() + buf)
    outfile.flush()
    return outfile


registerwriter('camxfiles.one3d', ncf2one3d)
registerwriter('one3d', ncf2one3d)



__doc__ = """
.. _point_source
:mod:`point_source` -- Point Source File Interfaces
===================================================

.. module:: point_source
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map and random access read
   based file interfaces for CAMx point source files.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmap', 'Read', 'Write']

from . import Memmap
from . import Read
from . import Write

if __name__ == '__main__':
    from PseudoNetCDF.camxfiles.point_source.Memmap import point_source
    from PseudoNetCDF.pncdump import pncdump_parser, \
        dump_from_cmd_line
    parser = pncdump_parser()
    (file_path, options, extra_args_dict) = parser.parse_args()

    dump_from_cmd_line(file_path, options, point_source)



#!/usr/bin/env python
from __future__ import print_function

import getpass
from urllib.parse import urlencode
from urllib.request import urlopen, Request

import time
from datetime import datetime
import os
import argparse

import numpy as np
from netCDF4 import Dataset


def getdate(x):
    return datetime.strptime(x, '%Y-%m-%d').strftime('%Y%m%d')


parser = argparse.ArgumentParser("AQS Mart for pnceval")
parser.add_argument('-u', '--username', required=True, type=str,
                    help="User name for EPA's AQS Data Mart \"Query Air Data\" application at https://ofmext.epa.gov/AQDMRS/aqdmrs.html")
parser.add_argument('-p', '--password', required=False, type=str,
                    help="Password for EPA's AQS Data Mart \"Query Air Data\" application at https://ofmext.epa.gov/AQDMRS/aqdmrs.html")
parser.add_argument('-s', '--start-date', required=True, dest='bdate',
                    type=getdate, help='Start date (inclusive) YYYYMMDD')
parser.add_argument('-e', '--end-date', required=True, dest='edate',
                    type=getdate, help='Start date (inclusive) YYYYMMDD')
parser.add_argument('--param', type=str, default='44201',
                    nargs='?', help="Must exist as an AQS parameter")
parser.add_argument('--frmonly', type=str, choices=[
                    'Y', 'N'], nargs='?', default='Y', help='Federal register method only')
parser.add_argument('-d', '--download-output', dest='downout', type=str,
                    nargs='?', default='AQSREST_DATA.csv', help='Path for downloaded output.')
parser.add_argument('-o', '--output', type=str, dest='outpath',
                    nargs='?', default='AQSREST_DATA.nc', help='Path for output.')
parser.add_argument('-O', '--overwrite', dest='overwrite', default=False,
                    action='store_true', help='Ovewrite if output already exists.')
spacegroup = parser.add_mutually_exclusive_group(required=True)
spacegroup.add_argument(
    '--gridcro2d', help='CMAQ MCIP GRIDCRO2D file or any file that has LAT and LON variables')
spacegroup.add_argument(
    '--bbox', help='minlon,minlat,maxlon,maxlat in degrees east and north or 0N, 0E')


args = parser.parse_args()

if args.password is None:
    args.password = getpass.getpass(
        prompt='AQS DATA Mart password: ', stream=None)

# authvalues = dict(username = args.username, password = args.password)
# auth = urlencode(authvalues).encode()

if args.gridcro2d is not None:
    f = Dataset(args.GRIDCRO2D)
    lon = f.variables['LON'][0, 0]
    lat = f.variables['LAT'][0, 0]
    args.minlon = llcrnrlon = lon[:, 0].max()
    args.maxlon = urcrnrlon = lon[:, -1].min()
    args.minlat = llcrnrlat = lat[0, :].max()
    args.maxlat = urcrnrlat = lat[-1, :].min()
if args.bbox is not None:
    args.minlon, args.minlat, args.maxlon, args.maxlat = [
        float(c) for c in args.bbox.split(',')]

print(args.minlon, args.maxlon)
print(args.minlat, args.maxlat)

if os.path.exists(args.outpath) and not args.overwrite:
    raise IOError('Path already exists: %s' % args.outpath)


def getrest():
    url = "https://aqs.epa.gov/api/rawDataNotify?format=DMCSV&user=%(username)s&password=%(password)s&param=%(param)s&bdate=%(bdate)s&edate=%(edate)s&minlat=%(minlat)s&maxlat=%(maxlat)s&minlon=%(minlon)s&maxlon=%(maxlon)s&dur=1&frmonly=%(frmonly)s" % dict(
        args._get_kwargs())
    urlrequest = Request(url)  # , data = auth)

    print(urlrequest, type(urlrequest))
    import pdb
    pdb.set_trace()
    dataid = urlopen(urlrequest).read()
    print(dataid)
    urlstatus = urlencode('https://aqs.epa.gov/api/status?id=' + dataid)
    while True:
        status = urlopen(urlstatus).read().strip()
        print(status)
        print(status, dataid)
        if status in ('Submitted', 'Processing'):
            time.sleep(3)
        elif status == 'Error':
            raise ValueError('AQS Mart did not correctly process your result.')
        elif status == 'Completed':
            break
        else:
            raise ValueError(
                'AQS Mart returned unknown status (%s). Expected Submitted, Processing, Completed or Error.')
    urldata = 'https://aqs.epa.gov/api/retrieve?id=' + dataid
    stmt = ('wget --no-check-certificate -O %s "%s"' % (args.downout, urldata))
    os.system(stmt)


getrest()


def hourly_parser(*args):
    # import dateutil
    # parse = np.vectorize(dateutil.parser.parse)
    out = (args[0] + ' ' + args[1].astype('S2') + 'Z').astype(np.datetime64)
    return out


try:
    import pandas
except Exception:
    raise ImportError(
        'pncaqsrest4pnceval.py requires pandas; install pandas (e.g., pip install pandas)')
data = pandas.read_csv(args.downout, skip_footer=2, parse_dates=[[12, 13]], date_parser=hourly_parser, converters={
                       'State Code': str, 'County Code': str, 'Site Num': str}, index_col=False, engine='python')

hourly = data.groupby(['Parameter Code', 'AQS Parameter Desc', 'Unit of Measure', 'Sample Duration', 'Date GMT_24 Hour GMT', 'State Code', 'County Code', 'Site Num'], as_index=False).aggregate(
    np.mean).sort(['Parameter Code', 'AQS Parameter Desc', 'Unit of Measure', 'Sample Duration', 'Date GMT_24 Hour GMT', 'State Code', 'County Code', 'Site Num'])
sites = data.groupby(['State Code', 'County Code', 'Site Num'],
                     as_index=False).aggregate(np.mean)
nsites = len(sites)

outf = Dataset(args.outpath, 'w')
outf.createDimension('time', None)
outf.createDimension('LAY', 1)
sitelist = [row['State Code'] + row['County Code'] + row['Site Num']
            for idx, row in sites.iterrows()]
outf.SITENAMES = ';'.join(sitelist)

outf.createDimension('points', len(sitelist))
outf.lonlatcoords = '/'.join(['%f,%f' % (row['Longitude'], row['Latitude'])
                              for idx, row in sites.iterrows()])
last_time = start_time = hourly['Date GMT_24 Hour GMT'].values.min()
end_time = hourly['Date GMT_24 Hour GMT'].values.max()
ntimes = np.timedelta64(end_time - start_time, 'h').astype('i') + 1
alltimes = [np.timedelta64(i, 'h') + start_time for i in range(ntimes)]
temp = {}
lat = outf.createVariable('latitude', 'f', ('points',))
lat.units = 'degrees_north'
lat.standard_name = 'latitude'
lat[:] = sites['Latitude'].values
lon = outf.createVariable('longitude', 'f', ('points',))
lon.units = 'degrees_east'
lon.standard_name = 'longitude'
lon[:] = sites['Longitude'].values

for idx, row in hourly.iterrows():
    this_time = row['Date GMT_24 Hour GMT']
    val = row['Sample Measurement']
    unit = row['Unit of Measure'].strip()
    aqsid = row['State Code'] + row['County Code'] + row['Site Num']
    sidx = sitelist.index(aqsid)
    var_name = row['AQS Parameter Desc']
    if var_name not in outf.variables.keys():
        var = outf.createVariable(var_name, 'f', ('time', 'LAY', 'points'))
        var.units = unit
        var.standard_name = var_name
        temp[var_name] = np.zeros((ntimes, 1, nsites), dtype='f')
    tmpvar = temp[var_name]
    var = outf.variables[var_name]
    assert(var.units == unit)
    if last_time != this_time:
        last_time = this_time
        tidx = alltimes.index(np.datetime64(this_time))
    if tidx > ntimes:
        raise ValueError('Times (%d) exceed expected (%d)' % (tidx, ntimes))

    tmpvar[tidx, 0, sidx] = val

for varkey, tempvals in temp.items():
    outf.variables[varkey][:] = tempvals



__all__ = ['wind']
__doc__ = """
.. _Read
:mod:`Read` -- wind Read interface
============================================

.. module:: Read
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` random access read for CAMx
              wind files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timediff, timeadd, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import OpenRecordFile, read_into
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariables


# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class wind(PseudoNetCDFFile):
    """
    wind provides a PseudoNetCDF interface for CAMx
    wind files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> wind_path = 'camx_wind.bin'
        >>> rows,cols = 65,83
        >>> windfile = wind(wind_path,rows,cols)
        >>> windfile.variables.keys()
        ['TFLAG', 'U', 'V']
        >>> v = windfile.variables['V']
        >>> tflag = windfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> windfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    time_hdr_fmts = {12: "fii", 8: "fi"}
    data_fmt = "f"

    def __init__(self, rf, rows=None, cols=None):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """
        self.rffile = OpenRecordFile(rf)
        self.time_hdr_fmt = self.time_hdr_fmts[self.rffile.record_size]

        self.time_hdr_size = struct.calcsize(self.time_hdr_fmt)
        self.padded_time_hdr_size = struct.calcsize("ii" + self.time_hdr_fmt)
        self.__readheader()
        self.__gettimestep()
        if rows is None and cols is None:
            rows = self.cell_count
            cols = 1
        elif rows is None:
            rows = self.cell_count / cols
        elif cols is None:
            cols = self.cell_count / rows
        else:
            if cols * rows != self.cell_count:
                raise ValueError(("The product of cols (%d) and rows (%d) " +
                                  "must equal cells (%d)") %
                                 (cols, rows, self.cell_count))
        self.createDimension('TSTEP', self.time_step_count)
        self.createDimension('COL', cols)
        self.createDimension('ROW', rows)
        self.createDimension('LAY', self.nlayers)

        self.variables = PseudoNetCDFVariables(self.__var_get, ['U', 'V'])

    def __var_get(self, key):
        def constr(uv):
            return self.getArray()[
                :, {'U': 0, 'V': 1}[uv], :, :, :].copy()

        def decor(uv):
            return dict(units='m/s',
                        var_desc=uv.ljust(16),
                        long_name=uv.ljust(16))
        values = constr(key)

        var = self.createVariable(key, 'f', ('TSTEP', 'LAY', 'ROW', 'COL'))
        var[:] = values
        for k, v in decor(key).items():
            setattr(var, k, v)
        return var

    def __readheader(self):
        """
        __readheader reads the header section of the ipr file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """
        self.data_start_byte = 0
        if self.time_hdr_fmt == 'fii':
            self.start_time, self.start_date, self.lstagger = self.rffile.read(
                self.time_hdr_fmt)
        elif self.time_hdr_fmt == 'fi':
            self.start_time, self.start_date = self.rffile.read(
                self.time_hdr_fmt)
            self.lstagger = None
        else:
            raise NotImplementedError("Header format is unknown")

        self.record_size = self.rffile.record_size
        self.padded_size = self.record_size + 8
        self.cell_count = self.record_size // struct.calcsize(self.data_fmt)
        self.record_fmt = self.data_fmt * self.cell_count

    def __gettimestep(self):
        """
        Header information provides start and end date, but does not
        indicate the increment between.  This routine reads the first
        and second date/time and initializes variables indicating the
        timestep length and the anticipated number.
        """
        # This is a bit of a hack, but should work:
        # Search for the next record that is the same
        # length as self.padded_time_hdr_size
        #
        # This should be the next date record
        # 1) date - startdate = timestep
        # 2) (record_start - self.padded_time_hdr_size)/self.padded_size
        #     = klayers
        self.rffile._newrecord(0)
        self.rffile.next()
        nlayers = 0
        while not self.rffile.record_size == self.time_hdr_size:
            self.rffile.next()
            nlayers += 1

        self.nlayers = (nlayers - 1) // 2

        if self.time_hdr_fmt == "fi":
            time, date = self.rffile.read(self.time_hdr_fmt)
        elif self.time_hdr_fmt == "fii":
            time, date, lstagger = self.rffile.read(self.time_hdr_fmt)

        self.end_time, self.end_date = time, date

        self.time_step = timediff(
            (self.start_date, self.start_time), (date, time))

        while True:
            try:
                for i in range(self.nlayers * 2 + 1):
                    self.rffile.next()
                if self.rffile.record_size == 8:
                    self.end_time, self.end_date = self.rffile.read("fi")
                elif self.rffile.record_size == 12:
                    self.end_time, self.end_date, lstagger = self.rffile.read(
                        "fii")
                else:
                    raise KeyError()
            except Exception:
                break

        self.time_step_count = int(timediff((self.start_date, self.start_time),
                                            (self.end_date, self.end_time)) /
                                   self.time_step) + 1

    def __layerrecords(self, k):
        return k - 1

    def __timerecords(self, dt):
        """
        routine returns the number of records to increment from the
        data start byte to find the first time
        """
        d, t = dt
        nsteps = int(
            timediff((self.start_date, self.start_time),
                     (d, t)) / self.time_step)
        nlays = self.__layerrecords(self.nlayers + 1)
        return nsteps * nlays

    def __recordposition(self, date, time, k, duv):
        """
        routine uses pagridrecords, timerecords,irecords,
        jrecords, and krecords multiplied by the fortran padded size
        to return the byte position of the specified record

        pagrid - integer
        date - integer
        time - float
        duv - integer (0=date,1=uwind,2=vwind)
        """
        bytes = self.data_start_byte
        nsteps = self.__timerecords((date, time))
        bytes += int(nsteps / self.nlayers) * self.padded_time_hdr_size
        bytes += int(nsteps / self.nlayers) * 12
        bytes += nsteps * self.padded_size * 2
        if not duv == 0:
            bytes += self.padded_time_hdr_size
            bytes += self.__layerrecords(k) * 2 * self.padded_size
        if duv == 2:
            bytes += self.padded_size
        return bytes

    def seek(self, date=None, time=None, k=1, uv=1):
        """
        Move file cursor to beginning of specified record
        see __recordposition for a definition of variables
        """
        if date is None:
            date = self.start_date
        if time is None:
            time = self.start_time
        chkvar = True
        if (
            chkvar and
            timediff((self.end_date, self.end_time), (date, time)) > 0 or
            timediff((self.start_date, self.start_time), (date, time)) < 0
        ):
            raise KeyError(("Wind file includes (%i,%6.1f) thru (%i,%6.1f); " +
                            "you requested (%i,%6.1f)") %
                           (self.start_date, self.start_time,
                            self.end_date, self.end_time, date, time))
        if chkvar and uv < 0 or uv > 2:
            raise KeyError("Wind file includes Date (uv: 0), u velocity " +
                           "(uv: 1) and v velocity (uv: 2); you requested %i" %
                           (uv))

        self.rffile._newrecord(self.__recordposition(date, time, k, uv))

    def read(self):
        """
        provide direct access to the underlying RecordFile read
        method
        """
        return self.rffile.read(self.record_fmt)

    def read_into(self, dest):
        """
        put values from rffile read into dest
        dest - numpy or numeric array
        """
        return read_into(self.rffile, dest, "", self.data_fmt)

    def seekandreadinto(self, dest, date=None, time=None, k=1, duv=1):
        """
        see seek and read_into
        """
        self.seek(date, time, k, duv)
        self.read_into(dest)

    def seekandread(self, date=None, time=None, k=1, duv=1):
        """
        see seek and read
        """
        self.seek(date, time, k, duv)
        return self.read()

    def keys(self):
        for d, t in timerange((self.start_date, self.start_time),
                              timeadd((self.end_date, self.end_time),
                                      (0, self.time_step)),
                              self.time_step):
            for k in range(1, self.nlayers + 1):
                yield d, t, k

    def values(self):
        for d, t, k in self.keys():
            yield self.seekandread(d, t, k, 1), self.seekandread(d, t, k, 2)

    def items(self):
        for d, t, k in self.keys():
            y1, y2 = self.seekandread(d, t, k, 1), self.seekandread(d, t, k, 2)
            yield d, t, k, y1, y2

    __iter__ = keys

    def getArray(self, krange=slice(1, None)):
        if type(krange) != slice:
            if type(krange) == tuple:
                krange = slice(*krange)
            if type(krange) == int:
                krange = slice(krange, krange + 1)
        a = zeros(
            (
                self.time_step_count,
                2,
                len(range(*krange.indices(self.nlayers + 1))),
                len(self.dimensions['ROW']),
                len(self.dimensions['COL']),
            ), 'f')
        nlay = self.nlayers
        for i, (d, t) in enumerate(self.timerange()):
            for uv in range(1, 3):
                for ki, k in enumerate(range(*krange.indices(nlay + 1))):
                    uvi = uv - 1
                    ki = k - 1
                    self.seekandreadinto(a[i, uvi, ki, :, :], d, t, k, uv)
        return a

    def timerange(self):
        return timerange((self.start_date, self.start_time),
                         timeadd((self.end_date, self.end_time),
                                 (0, self.time_step)),
                         self.time_step)


class TestRead(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testWD(self):
        import PseudoNetCDF.testcase
        wdfile = wind(PseudoNetCDF.testcase.camxfiles_paths['wind'], 4, 5)
        checkv = array([
            -1.73236704e+00, -1.99612117e+00, -3.00912833e+00, -3.92667437e+00,
            -3.49521232e+00, 2.04542422e+00, 8.57666790e-01, -1.71201074e+00,
            -4.24386787e+00, -5.37704515e+00, 1.85697508e+00, 6.34313405e-01,
            -1.21529281e+00, -3.03180861e+00, -4.36278439e+00, -1.90753967e-01,
            -1.08261776e+00, -1.73634803e+00, -2.10829663e+00, -2.28424144e+00,
            -1.88443780e+00, -2.02582169e+00, -3.09955978e+00, -4.14587784e+00,
            -3.72402787e+00, 2.16277528e+00, 8.94082963e-01, -1.86343944e+00,
            -4.58147812e+00, -5.81837606e+00, 1.97949493e+00, 6.12511635e-01,
            -1.35096896e+00, -3.25313163e+00, -4.67790413e+00, -1.89851984e-01,
            -1.16381800e+00, -1.84269297e+00, -2.21348834e+00, -2.40952253e+00,
            -2.04972148e+00, -2.11795568e+00, -3.06094027e+00, -4.11207581e+00,
            -3.74964952e+00, 2.09780049e+00, 8.01259458e-01, -1.90404522e+00,
            -4.59170580e+00, -5.83114100e+00, 1.97475648e+00, 5.54396451e-01,
            -1.41695607e+00, -3.28227353e+00, -4.67724609e+00, -1.94723800e-01,
            -1.18353117e+00, -1.86556363e+00, -2.22842574e+00, -2.42080784e+00,
            -1.65720737e+00, -1.58054411e+00, -2.25336742e+00, -3.06462526e+00,
            -2.47261453e+00, 1.37642264e+00, 1.16142654e+00, -6.82058990e-01,
            -2.68112469e+00, -3.38680530e+00, 1.80796599e+00, 1.48641026e+00,
            -1.71508826e-02, -1.68607295e+00, -2.89399385e+00, 3.40398103e-01,
            3.25049832e-02, -5.91206312e-01, -1.19038010e+00, -1.52301860e+00,
            -1.83006203e+00, -1.74505961e+00, -2.50190806e+00, -3.29507184e+00,
            -2.65367699e+00, 1.55719578e+00, 1.25234461e+00, -9.14537191e-01,
            -3.16307521e+00, -4.00584650e+00, 2.07018161e+00, 1.60957754e+00,
            -1.46312386e-01, -2.04018188e+00, -3.40377665e+00, 4.11731720e-01,
            -2.29119677e-02, -7.27373540e-01, -1.35116744e+00, -1.70711970e+00,
            -1.72859466e+00, -1.73683071e+00, -2.65253377e+00, -3.43689489e+00,
            -2.75470304e+00, 1.58366191e+00, 1.19656324e+00, -1.09935236e+00,
            -3.38544369e+00, -4.26436615e+00, 2.04826832e+00, 1.53576791e+00,
            -2.60809243e-01, -2.18679833e+00, -3.59082842e+00, 3.77060443e-01,
            -1.05680525e-01, -8.10511589e-01, -1.40993130e+00,
            -1.76300752e+00], dtype='f').reshape(2, 3, 4, 5)
        self.assertTrue((wdfile.variables['V'][:] == checkv).all())


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function, unicode_literals
__all__ = ['platform_is_bigendian', 'freadnumpy', 'freadstruct', 'fread',
           'needs_byteswap', 'check_read', 'RecordFile', 'unpack_from_file',
           'seek_to_record', 'read_into', 'writeline', 'OpenRecordFile',
           'Int2Asc', 'Asc2Int']

__doc__ = """
.. _FortranFileUtil
:mod:`FortranFileUtil` -- Fortran File interaction functions
============================================================

.. module:: FortranFileUtil
   :platform: Unix, Windows
   :synopsis: Provides utilities for interacting with Fortran
              "unformatted" binary files.  These files often
              have records with integer buffers and formatted data.
              Often the data has a byte order different than the
              machine doing the reading. These utilities provide ways
              for the user to get data and ignore the byte order and
              data buffers.
.. moduleauthor:: Adam Hupp
"""

# Distribution Packages
import unittest
import sys
import struct
from operator import mul
# Site-Packages
# use numpy if available
# otherwise use numeric
from numpy import array, fromfile
import numpy as np


platform_is_bigendian = (sys.byteorder != 'little')


def freadnumpy(ifile, count, read_type, return_type=None, byteswap=False):
    if return_type is None:
        return_type = read_type

    if byteswap ^ platform_is_bigendian:
        endian = '>'
    else:
        endian = '<'
    # rfmt = "%s%s" % (endian, return_type)
    fmt = "%s%s" % (endian, read_type)
    return array(fromfile(ifile, dtype=fmt, count=count), dtype=return_type)


def freadstruct(ifile, count, read_type, return_type=None, byteswap=False):
    if return_type is None:
        return_type = read_type

    if byteswap ^ platform_is_bigendian:
        endian = '>'
    else:
        endian = '<'

    fmt = "%s%d%s" % (endian, count, read_type)

    elem = unpack_from_file(fmt, ifile)

    return array(elem, return_type)


# numpy style fread is more than 2 x as fast as struct
# preserving freadstruct (previously fread) until further notice
fread = freadnumpy


def needs_byteswap(bigendian):
    """Determines if byteswapping is necessary for the given endianness
    For example, if we are reading bigendian data, call

    need_byteswap(True)

    to see if that data needs to be swapped on the current platform
    """

    return bigendian != platform_is_bigendian


def check_read(requested, result):
    """Checks that the result of a read from a file was of the expected size
    Raises EOFError if nothing was read, or IOError if the read was incomplete
    """
    if len(result) != requested and requested != 0:
        if result == '':
            raise EOFError()
        else:
            raise IOError("Incomplete read, requested %d recieved %d" %
                          (requested, len(result)))


class RecordFile(object):
    """
    This class provides an easy interface to treat unformated
    Fortran files like like record files.

    TODO: count=-1 does not work in aread, off by one element?!?
    TODO: error checking in aread, prevent reading past end of record
    """

    def tell(self):
        """
        use file object tell
        """
        return self.infile.tell()

    def __init__(self, infile, bigendian=True):
        """
        Arguments:
        infile --  an open file-like object.  Must be random access.  Must
          be a real file if the RecordFile.aread method is going to be used.
        bigendian --  boolean, True if file is bigendian

        ***Assumes single 4 byte integer pad on either end
        """
        if hasattr(infile, 'seek') and \
           hasattr(infile, 'tell') and \
           hasattr(infile, 'read'):
            self.infile = infile
        else:
            try:
                self.infile = open(infile, 'rb')
            except Exception as e:
                raise TypeError("infile is of " + str(type(infile)) +
                                ": and can only be string or file; " + str(e))

        # Is this platform a different byteorder than the file?
        self.byteswap = needs_byteswap(bigendian)

        if bigendian:
            self.format_prefix = '>'
        else:
            self.format_prefix = '<'

        self.infile.seek(0, 2)
        self.length = self.infile.tell()

        self._newrecord(0)

    def seek(self, offset):
        if offset < self.length:
            posnow = self.infile.tell()
            relmov = offset - posnow
            self.infile.seek(relmov, 1)
        else:
            raise ValueError(("File length = %i; record start " +
                              "requested %f") % (self.length, offset))

    def _newrecord(self, offset):
        """Move to a new record starting at the given offset
        """
        self.seek(offset)
        self.record_start = offset
        self.record_size = self.unpack("i")[0]

    def unpack(self, fmt):
        """Unpack a set of values, like the struct module
        """
        data = unpack_from_file(self.format_prefix + fmt, self.infile)
        data = tuple([d.decode() if isinstance(d, bytes) else d for d in data])
        return data

    def read(self, fmt):
        """Unpack a set of values, like the struct module
        """
        result = unpack_from_file(self.format_prefix + fmt, self.infile)
        self.next()
        return result

    def next(self):
        """Move to the next record.
        Returns True if sucessful, False if no more records available.
        """

        # 4 bytes for record size, start and end
        offset = self.record_start + self.record_size + 8
        if offset < self.length:
            self._newrecord(offset)
            return True
        else:
            self.infile.seek(0, 2)  # EOF
            return False

    def previous(self):
        """Move to the previous record.
        Returns True if sucessful, False if no more records available.
        """

        # 4 bytes for record size of this record
        # 4 bytes for record size of previous record
        if self.tell() == 4:
            return False
        offset = (-8, -4)[self.eof()]

        # move to size byte of previous record
        self.infile.seek(offset, 1)

        # Read size of previous record
        offset = self.unpack("i")[0]

        # actual offset = this  record start - size -
        #                 previous record length + header
        offset = self.tell() - offset - 8
        if offset >= 0:
            self._newrecord(offset)
            return True
        else:
            self.infile.seek(0, 1)  # EOF
            return False

    def aread(self, type, count):
        """Read a Numeric array

        Arguments:
        type -- a type code, one of those used by the struct or Numeric modules
        count -- number of elements to read.  If not specified,
                  defaults to all remaining bytes in current record
        """

        if count == -1:
            # NOTE: This does not work right!
            remain = self.record_size - (self.infile.tell() -
                                         self.record_start)
            count = remain / struct.calcsize(type)

        data = fread(self.infile, count,
                     type, type,
                     self.byteswap)
        data = np.array([d.decode() if hasattr(
            d, 'decode') else d for d in data])
        return data

    def eof(self):
        """Returns true if this RecordFile is at the end of the file
        """
        return self.infile.tell() == self.length

    def restart_record(self):
        """Move to beginning of record
        """
        self._newrecord(self.record_start)


def unpack_from_file(fmt, filein):
    """Like struct.unpack, but reads from a file instead of a string
    """
    fmtsize = struct.calcsize(fmt)
    data = filein.read(fmtsize)
    check_read(fmtsize, data)
    data = struct.unpack(fmt, data)
    data = tuple([d.decode() if hasattr(d, 'decode') else d for d in data])
    return data


def seek_to_record(rf, rid, fmt):
    """Searches for a record beginning with rid by unpacking
    the first struct.calcsize(fmt) bytes and comparing the
    results
    """
    while True:
        cid = rf.unpack(fmt)
        if rid == cid:
            rf.restart_record()
            break
        else:
            if not rf.next():
                raise ValueError("Time %s not found" % str(rid))


def read_into(rf, dest, id_fmt, data_fmt='f'):
    """Read an array from a RecordFile into a Numeric array.
    I don't know how this will work if dest is anything but 2D.

    Arguments:
    rf -- a RecordFile instance
    dest -- a Numeric array, possibly a slice.

    There must be at least product(shape(dest)) elements
    left in the current record

    """
    from functools import reduce
    if rf.eof():
        raise EOFError()

    id = rf.unpack(id_fmt)
    rd = rf.aread(data_fmt, reduce(mul, dest.shape))
    dest[...] = rd.reshape(dest.shape)
    return id


def writeline(d, fmt, ForceBig=True):
    """writeline appends length integers
    and determines if byteswap is necessary
    """
    rlen = struct.calcsize(fmt)
    rfmt = "i" + fmt + "i"
    if sys.byteorder == 'little' or ForceBig:
        rfmt = '>' + rfmt
    d = [i for i in d]
    d.insert(0, rlen)
    d.append(rlen)
    try:
        return struct.pack(rfmt, *d)
    except Exception:
        print(d)
        raise


def OpenRecordFile(rf):
    """All CAMx files use FortranFileUtil.RecordFiles
    as inputs.  This function decides if the input was
    of the right type.  If not, it tries to make a RecordFile
    from the input.

    rf - str, unicode, file, RecordFile
    """
    if type(rf) == RecordFile:
        pass
    else:
        rf = RecordFile(rf)
    rf._newrecord(0)
    return rf


def Int2Asc(mspec):
    """Some CAMx input files have text stored as
    integers.  This function helps to undo that
    """
    spcname = ""
    for c in mspec:
        spcname += chr((((c - 32) // 256 - 32) // 256 - 32) // 256)
    return spcname


def Asc2Int(spcname):
    """Some CAMx output files need the text stored
    as integers.  This function helps to do that
    """
    mspec = []
    for c in spcname:
        mspec.append(int((((((ord(c) * 256) + 32) * 256) + 32) * 256) + 32))
    return mspec


class TestFileUtils(unittest.TestCase):
    def setUp(self):
        from tempfile import TemporaryFile as tf
        self.tmpfile = tf(mode='w+b')

        # writing tempfile with Fortran unformatted binary format
        # 1st line is 0-19 as floats
        # 2nd line is 0-19 as ints
        # 3rd line is 0-19 as strings
        write = self.tmpfile.write
        write(b'\x00\x00\x00P' +
              b'\x00\x00\x00\x00?\x80\x00\x00@\x00' +
              b'\x00\x00@@\x00\x00@\x80\x00\x00@\xa0' +
              b'\x00\x00@\xc0\x00\x00@\xe0\x00\x00A' +
              b'\x00\x00\x00A\x10\x00\x00A \x00\x00A0' +
              b'\x00\x00A@\x00\x00AP\x00\x00A`\x00\x00Ap' +
              b'\x00\x00A\x80\x00\x00A\x88\x00\x00A' +
              b'\x90\x00\x00A\x98\x00\x00' + b'\x00\x00\x00P')
        write(b'\x00\x00\x00P' + b'\x00\x00\x00\x00\x00\x00\x00\x01' +
              b'\x00\x00\x00\x02\x00\x00\x00\x03' +
              b'\x00\x00\x00\x04\x00\x00\x00\x05' +
              b'\x00\x00\x00\x06\x00\x00\x00\x07' +
              b'\x00\x00\x00\x08\x00\x00\x00\t' +
              b'\x00\x00\x00\n\x00\x00\x00\x0b\x00\x00\x00\x0c' +
              b'\x00\x00\x00\r\x00\x00\x00\x0e\x00\x00\x00\x0f' +
              b'\x00\x00\x00\x10\x00\x00\x00\x11' +
              b'\x00\x00\x00\x12\x00\x00\x00\x13' + b'\x00\x00\x00P')
        write(b'\x00\x00\x00\x14' +
              b'The quick brown fox ' + b'\x00\x00\x00\x14')
        self.tmprf = RecordFile(self.tmpfile)

    def testAdvance(self):
        self.tmprf._newrecord(0)
        self.assertEquals(self.tmprf.tell(), 4)
        self.tmprf.next()
        self.assertEquals(self.tmprf.tell(), 92)
        self.tmprf.next()
        self.assertEquals(self.tmprf.tell(), 180)
        self.failIf(self.tmprf.next())
        self.assertEquals(self.tmprf.tell(), 204)
        self.failIf(self.tmprf.next())
        self.tmprf.previous()
        self.assertEquals(self.tmprf.tell(), 180)
        self.tmprf.previous()
        self.assertEquals(self.tmprf.tell(), 92)
        self.tmprf.previous()
        self.assertEquals(self.tmprf.tell(), 4)
        self.failIf(self.tmprf.previous())
        self.assertEquals(self.tmprf.tell(), 4)

    def testFloat(self):
        from numpy import arange
        self.tmprf._newrecord(0)
        self.assertTrue((arange(20, dtype='f') ==
                        self.tmprf.aread('f', 20)).all())

    def testReadInto(self):
        from numpy import arange, zeros
        dest = zeros((4, 5), 'f')

        self.tmprf._newrecord(0)

        # Necessary because written to anticipate fortran swapping of axes
        read_into(self.tmprf, dest, '', 'f')
        self.assertTrue((arange(20, dtype='f').reshape((4, 5)) == dest).all())

    def testInt(self):
        from numpy import arange
        self.tmprf._newrecord(0)
        self.tmprf.next()
        self.assertTrue((arange(20, dtype='i') ==
                        self.tmprf.aread('i', 20)).all())

    def testSeek(self):
        self.tmprf._newrecord(0)
        seek_to_record(self.tmprf, (0., 1., 2.), "fff")
        self.assertEquals(self.tmprf.tell(), 4)
        seek_to_record(self.tmprf, (0, 1, 2), "iii")
        self.assertEquals(self.tmprf.tell(), 92)
        seek_to_record(self.tmprf, ('The',), '3s')
        self.assertEquals(self.tmprf.tell(), 180)
        seek_to_record(self.tmprf, ('T', 'h', 'e',), '3c')
        self.assertEquals(self.tmprf.tell(), 180)

    def testStr(self):
        from numpy import array
        self.tmprf._newrecord(0)
        self.tmprf.next()
        self.tmprf.next()
        checkv = array(["The quick brown fox "])
        testv = self.tmprf.aread('S20', 1)
        self.assertTrue(np.any(checkv == testv))

    def runTest(self):
        pass


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict

__doc__ = r"""
.. _units
:mod:`units` -- Functions for converting units
========================================================

.. module:: units
   :platform: Unix, Windows
   :synopsis: Provides functions for converting units.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

__all__ = ['F2C', 'F2K', 'K2C', 'K2F', 'KCMAQ2F', 'M2km', 'MPS2kph',
           'convert', 'converter', 'converters_dict', 'km2m', 'm2ft',
           'm2km', 'm2miles', 'min2h', 'molespsCMAQ2molesph', 'mps2kmps',
           'mps2kph', 'mps2milesph', 'mps2milesps', 's2h', 's2min']


def s2min(a):
    return a / 60.


def min2h(a):
    return a / 60.


def s2h(a):
    return min2h(s2min(a))


def m2km(a):
    return a * .001


def M2km(a):
    return a * .001


def km2m(a):
    return a * 1000.


def m2ft(a):
    return a * 3.2808399


def m2miles(a):
    return a * 0.000621371192


def K2C(a):
    return a - 273.2


def K2F(a):
    return K2C(a) * 9 / 5. + 32


def KCMAQ2F(a):
    return K2C(a) * 9 / 5. + 32


def F2C(a):
    return (a - 32.) * 5 / 9.


def F2K(a):
    return F2C(a) + 273.2


def mps2kph(a):
    return a * 3.6


def MPS2kph(a):
    return a * 3.6


def mps2kmps(a):
    return a * .001


def mps2milesph(a):
    return a * 2.23693629


def mps2milesps(a):
    return a * 0.000621371192


def molespsCMAQ2molesph(a):
    return a * 60 * 60


def ppm2ppb(a):
    return a * 1000.


def ppb2ppt(a):
    return a * 1000.


def ppt2ppb(a):
    return a / 1000.


def ppb2ppm(a):
    return a / 1000.


def ppm2ppt(a):
    return ppm2ppb(ppb2ppt(a))


def ppt2ppm(a):
    return ppt2ppb(ppb2ppm(a))


class converters_dict(defaultdict):
    def __init__(self, dct):
        for k, v in dct.items():
            self[k] = v

    def __missing__(self, key):
        if type(key) == tuple and key[0] == key[1]:
            return lambda a: a


converter = converters_dict({
    ('s', 'min'): s2min,
    ('min', 'h'): min2h,
    ('s', 'h'): s2h,
    ('m/s', 'km/h'): mps2kph,
    ('M/S             ', 'km/h'): MPS2kph,
    ('m/s', 'miles/h'): mps2milesph,
    ('m/s', 'miles/s'): mps2milesps,
    ('m', 'km'): m2km,
    ('M               ', 'km'): M2km,
    ('km', 'm'): km2m,
    ('m', 'ft'): m2ft,
    ('m', 'miles'): m2miles,
    ('moles/s         ', 'moles/h'): molespsCMAQ2molesph,
    ('F', 'K'): F2K,
    ('F', 'C'): F2C,
    ('K', 'C'): K2C,
    ('K', 'F'): K2F,
    ('K               ', 'deg_F'): KCMAQ2F,
    ('K', 'deg_F'): K2F,
    ('ppm', 'ppb'): ppm2ppb,
    ('ppb', 'ppt'): ppb2ppt,
    ('ppt', 'ppb'): ppt2ppb,
    ('ppb', 'ppm'): ppb2ppm,
    ('ppt', 'ppm'): ppt2ppm,
    ('ppm', 'ppt'): ppm2ppt})


def convert(var, inunit, outunit):
    """
    Function that converts var from inunit to outunit.

    both units must be in converter dictionary and be compatible
    """
    return converter[(inunit, outunit)](var)



__doc__ = """
.. _Writers
:mod:`Writers` -- CAMx Write Interfaces
========================================

.. module:: Writers
   :platform: Unix, Windows
   :synopsis: Provides an easy access point for memory map based
              :ref:`PseudoNetCDF` file interfaces for all CAMx
              related files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['ncf2cloud_rain',
           'ncf2height_pressure',
           'ncf2humidity',
           'ncf2landuse',
           'ncf2one3d',
           'ncf2point_source',
           'ncf2temperature',
           'ncf2uamiv',
           'ncf2vertical_diffusivity',
           'ncf2wind',
           'ncf2lateral_boundary']

from .uamiv.Write import ncf2uamiv
from .cloud_rain.Write import ncf2cloud_rain
from .height_pressure.Write import ncf2height_pressure
from .point_source.Write import ncf2point_source
from .humidity.Write import ncf2humidity
from .landuse.Write import ncf2landuse
from .lateral_boundary.Write import ncf2lateral_boundary
from .temperature.Write import ncf2temperature
from .wind.Write import ncf2wind
from .vertical_diffusivity.Write import ncf2vertical_diffusivity
from .one3d.Write import ncf2one3d



from __future__ import print_function
from PseudoNetCDF.sci_var import PseudoNetCDFFile
import numpy as np
from datetime import datetime, timedelta


def getbdate(x):
    if x is None:
        return x
    elif isinstance(x, (datetime, np.datetime64)):
        return x
    else:
        return datetime.strptime(x + ' 00:00', '%Y-%m-%d %H:%M')


def getedate(x):
    if x is None:
        return x
    elif isinstance(x, (datetime, np.datetime64)):
        return x
    else:
        return datetime.strptime(x + ' 23:59', '%Y-%m-%d %H:%M')


class aqsraw(PseudoNetCDFFile):
    def __init__(self, yearpath, timeresolution='hourly', bdate=None,
                 edate=None, rdate=datetime(1900, 1, 1), wktpolygon=None,
                 sampleval=None, verbose=0):
        """
        yearpath - path to csv file from AQS
        timeresolution - choices = ['daily', 'hourly'] default = 'hourly',
                         defaults to hourly'
        bdate - Limit output so that the date starts at YYYY-MM-DD
        edate - Limit output so that the date end YYYY-MM-DD (inclusive)
        rdate - datetime(1900, 1, 1), dest = 'rdate', type = getrdate,
                help = 'Reference date YYYYMMDD HH:MM:SS')
        wktpolygon - WKT Polygon (default: None) equivalent to
                     "POLYGON ((-180 -90, 180 -90, 180 90, -180 90, -180 -90))"
        sampleval - Defaults to "Sample Measurement" for hourly and "Arithmetic
                    Mean" for daily
        verbose - level of verbosity
        """
        try:
            import pandas
        except Exception:
            raise ImportError('aqsraw requires pandas; install pandas ' +
                              '(e.g., pip install pandas)')
        if wktpolygon is not None:
            from shapely.wkt import loads
            from shapely.prepared import prep
            bounds = loads(wktpolygon)
            pbounds = prep(bounds)
        else:
            from shapely.geometry import Point
        bdate = getbdate(bdate)
        edate = getedate(edate)
        nseconds = {'hourly': 3600, 'daily': 3600 * 24}[timeresolution]
        tunit = {'hourly': 'hours', 'daily': 'days'}[timeresolution]

        if sampleval is None:
            if timeresolution == 'hourly':
                sampleval = "Sample Measurement"
            elif timeresolution == 'daily':
                sampleval = "Arithmetic Mean"
            else:
                raise KeyError(sampleval + ' not appropriate sampleval value')
        hourlys = []
        for yearpath in [yearpath]:
            if verbose > 0:
                print('Reading', yearpath)
            if timeresolution == 'hourly':
                parse_dates = [[11, 12]]
                date_key = 'Date GMT_Time GMT'
            else:
                parse_dates = [11]
                date_key = 'Date Local'
            data = pandas.read_csv(yearpath,
                                   index_col=False,
                                   converters={'State Code': str,
                                               'County Code': str,
                                               'Site Num': str},
                                   parse_dates=parse_dates)
            intimes = np.array([True]).repeat(len(data), 0)
            # intimes = intimes & (data['Parameter Code'] == param)

            if bdate is not None:
                intimes = intimes & (data[date_key] >= bdate)
            if edate is not None:
                intimes = intimes & (data[date_key] < edate)

            if wktpolygon is None:
                inspace_and_time = intimes
            else:
                inspace_and_time = np.array([False]).repeat(len(data), 0)
                zll = zip(intimes,
                          data['Longitude'].values,
                          data['Latitude'].values)
                for idx, (intime, plon, plat) in enumerate(zll):
                    if intime:
                        inspace_and_time[idx] = pbounds.contains(
                            Point(plon, plat))
            mask = inspace_and_time
            groupkeys = ['Parameter Code', 'Parameter Name',
                         'Units of Measure', date_key,
                         'State Code', 'County Code',
                         'Site Num']

            hourly = data[mask].groupby(groupkeys, as_index=False)\
                               .aggregate(np.mean)\
                               .sort_values(by=groupkeys)

            hourlys.append(hourly)

        if verbose > 0:
            print('Concatenating files')
        if len(hourlys) > 1:
            hourly = pandas.concat(hourlys)
        else:
            hourly = hourlys[0]

        sites = hourly.groupby(['State Code', 'County Code', 'Site Num'],
                               as_index=False).aggregate(np.mean)
        nsites = len(sites)

        rawdates = hourly[date_key]
        if bdate is None:
            bdate = rawdates.min()
        if edate is None:
            edate = rawdates.max()
        ntimes = int((edate - bdate).total_seconds() // nseconds) + 1
        alltimes = [timedelta(**{tunit: i}) + bdate for i in range(ntimes)]

        if verbose > 0:
            print('Creating output file')
        tdim = self.createDimension('time', ntimes)
        tdim.setunlimited(True)
        self.createDimension('LAY', 1)
        sitelist = [row['State Code'] + row['County Code'] +
                    row['Site Num'] for idx, row in sites.iterrows()]
        self.SITENAMES = ';'.join(sitelist)

        self.createDimension('points', len(sitelist))
        lonlatcoordstrs = ['%f,%f' % (row['Longitude'], row['Latitude'])
                           for idx, row in sites.iterrows()]
        self.lonlatcoords = '/'.join(lonlatcoordstrs)
        last_time = hourly[date_key].values.min()
        # end_time = hourly[date_key].values.max()
        temp = {}
        lat = self.createVariable('latitude', 'f', ('points',))
        lat.units = 'degrees_north'
        lat.standard_name = 'latitude'
        lat[:] = sites['Latitude'].values
        lon = self.createVariable('longitude', 'f', ('points',))
        lon.units = 'degrees_east'
        lon.standard_name = 'longitude'
        lon[:] = sites['Longitude'].values

        if verbose > 0:
            print('Processing data rows')

        for idx, row in hourly.iterrows():
            this_time = row[date_key]
            val = row[sampleval]
            unit = row['Units of Measure'].strip()
            aqsid = row['State Code'] + row['County Code'] + row['Site Num']
            sidx = sitelist.index(aqsid)
            var_name = row['Parameter Name']
            if var_name not in self.variables.keys():
                var = self.createVariable(
                    var_name, 'f', ('time', 'LAY', 'points'), fill_value=-999)
                var.units = unit
                var.standard_name = var_name
                temp[var_name] = np.ma.masked_all(
                    (ntimes, 1, nsites), dtype='f')
                temp[var_name].set_fill_value(-999)
            tmpvar = temp[var_name]
            var = self.variables[var_name]
            assert(var.units == unit)
            if last_time != this_time:
                last_time = this_time
                tidx = alltimes.index(this_time.to_pydatetime())
                if verbose > 0:
                    print('\b\b\b\b\b\b %3d%%' %
                          int(tidx / float(ntimes) * 100), end='', flush=True)
            if tidx > ntimes:
                raise ValueError(
                    'Times (%d) exceed expected (%d)' % (tidx, ntimes))

            tmpvar[tidx, 0, sidx] = val
        if verbose > 0:
            print('')
        if verbose > 0:
            print('Writing to file')
        for varkey, tempvals in temp.items():
            self.variables[varkey][:] = tempvals

        outtimes = np.array(
            [(t - rdate).total_seconds() / nseconds for t in alltimes])
        time = self.createVariable('time', outtimes.dtype.char, ('time',))
        time.units = rdate.strftime(tunit + ' since %F')
        time.standard_name = 'time'
        time[:] = outtimes


if __name__ == '__main__':
    import sys
    f = aqsraw(sys.argv[1], bdate=datetime(
        2015, 1, 1), edate=datetime(2015, 2, 1))



import unittest
import numpy as np
from PseudoNetCDF import PseudoNetCDFFile, PseudoNetCDFVariables
from PseudoNetCDF import PseudoNetCDFVariable
from . import requires_basemap, requires_pyproj
from PseudoNetCDF.pncwarn import warn


np_all_close = np.testing.assert_allclose


class PseudoNetCDFFileTest(unittest.TestCase):
    def setUp(self):
        from datetime import datetime, timedelta
        self.testncf = self._makencf()
        self.mymeta = set([
            'time', 'time_bounds', 'latitude', 'longitude', 'latitude_bounds',
            'longitude_bounds', 'lambert_conformal_conic'
        ])
        self.myvars = self.mymeta.union(['O3'])
        self.mydims = ['TIME', 'LAY', 'ROW', 'COL', 'nv', 'tnv']
        rtime = datetime.strptime(
            '1970-01-01 00:00:00+0000', '%Y-%m-%d %H:%M:%S%z')
        self.mytimes = np.array([rtime + timedelta(hours=i)
                                 for i in range(24)])

    def _makencf(self):
        from numpy import arange
        tncf = PseudoNetCDFFile()

        tncf.createDimension('TSTEP', 24)
        tncf.createDimension('LAY', 4)
        tncf.createDimension('ROW', 5)
        tncf.createDimension('COL', 6)
        tncf.createDimension('nv', 4)
        tncf.createDimension('tnv', 2)
        tncf.str_one = '1'
        tncf.int_two = 2
        tncf.float_threeptfive = 3.5
        tncf.Conventions = 'CF-1.6'
        o3 = tncf.createVariable('O3', 'f', ('TSTEP', 'LAY', 'ROW', 'COL'))

        o3[:] = arange(24 * 4 * 5 * 6).reshape(24, 4, 5, 6)
        o3.units = 'ppbv'
        o3.grid_mapping = 'lambert_conformal_conic'
        time = tncf.createVariable('time', 'd', ('TSTEP',))
        time.long_name = 'time'
        time.units = 'hours since 1970-01-01 00:00:00+0000'
        time[:] = np.arange(24)

        timeb = tncf.createVariable('time_bounds', 'd', ('TSTEP', 'tnv'))
        timeb.long_name = 'time_bounds'
        timeb.units = 'hours since 1970-01-01 00:00:00+0000'
        timeb[:, 0] = np.arange(0, 24)
        timeb[:, 1] = np.arange(1, 25)

        crs = tncf.createVariable('lambert_conformal_conic', 'i', ())
        crs.grid_mapping_name = 'lambert_conformal_conic'
        crs.standard_parallel = np.array([30., 45.])
        crs.longitude_of_central_meridian = -97.
        crs.latitude_of_projection_origin = 40.
        crs.false_northing = 1620000.
        crs.false_easting = 2412000.
        crs.semi_major_axis = 6371000.
        crs.semi_minor_axis = 6371000.
        lat = tncf.createVariable('latitude', 'f', ('ROW', 'COL'))
        lat.long_name = 'latitude'
        lat.units = 'degrees_north'
        lon = tncf.createVariable('longitude', 'f', ('ROW', 'COL'))
        lon.long_name = 'longitude'
        lon.units = 'degrees_east'
        latb = tncf.createVariable(
            'latitude_bounds', 'f', ('ROW', 'COL', 'nv'))
        latb.long_name = 'latitude_bounds'
        latb.units = 'degrees_north'
        lonb = tncf.createVariable(
            'longitude_bounds', 'f', ('ROW', 'COL', 'nv'))
        lonb.long_name = 'longitude_bounds'
        lonb.units = 'degrees_east'
        lon[:] = [
            [-120.21161038333193, -120.21160114763147, -120.21159191193058,
             -120.21158267622918, -120.21157344052737, -120.21156420482505],
            [-120.21161271536134, -120.21160347966001, -120.21159424395826,
             -120.21158500825604, -120.21157577255335, -120.21156653685021],
            [-120.21161504739118, -120.21160581168901, -120.21159657598642,
             -120.21158734028334, -120.2115781045798, -120.2115688688758],
            [-120.21161737942151, -120.21160814371851, -120.21159890801503,
             -120.21158967231109, -120.21158043660672, -120.21157120090189],
            [-120.21161971145229, -120.21161047574842, -120.21160124004409,
             -120.21159200433934, -120.21158276863409, -120.21157353292838]]

        lat[:] = [
            [22.748507533242535, 22.748509683865187, 22.74851183448702,
             22.74851398510802, 22.748516135728206, 22.748518286347593],
            [22.74851605050742, 22.748518201130356, 22.748520351752475,
             22.74852250237377, 22.74852465299425, 22.748526803613903],
            [22.74852456777256, 22.748526718395773, 22.748528869018187,
             22.748531019639763, 22.748533170260536, 22.74853532088048],
            [22.748533085037966, 22.74853523566145, 22.74853738628417,
             22.748539536906023, 22.7485416875271, 22.748543838147327],
            [22.74854160230359, 22.74854375292739, 22.748545903550376,
             22.748548054172538, 22.748550204793883, 22.7485523554144]]
        lonb[:] = [
            [[-120.21161038333193, -120.21160114763147,
              -120.21160347966001, -120.21161271536134],
             [-120.21160114763147, -120.21159191193058,
              -120.21159424395826, -120.21160347966001],
             [-120.21159191193058, -120.21158267622918,
              -120.21158500825604, -120.21159424395826],
             [-120.21158267622918, -120.21157344052737,
              -120.21157577255335, -120.21158500825604],
             [-120.21157344052737, -120.21156420482505,
              -120.21156653685021, -120.21157577255335],
             [-120.21156420482505, -120.2115549691223,
              -120.2115573011466, -120.21156653685021]],
            [[-120.21161271536134, -120.21160347966001,
              -120.21160581168901, -120.21161504739118],
             [-120.21160347966001, -120.21159424395826,
              -120.21159657598642, -120.21160581168901],
             [-120.21159424395826, -120.21158500825604,
              -120.21158734028334, -120.21159657598642],
             [-120.21158500825604, -120.21157577255335,
              -120.2115781045798, -120.21158734028334],
             [-120.21157577255335, -120.21156653685021,
              -120.2115688688758, -120.2115781045798],
             [-120.21156653685021, -120.2115573011466,
              -120.21155963317135, -120.2115688688758]],
            [[-120.21161504739118, -120.21160581168901,
              -120.21160814371851, -120.21161737942151],
             [-120.21160581168901, -120.21159657598642,
              -120.21159890801503, -120.21160814371851],
             [-120.21159657598642, -120.21158734028334,
              -120.21158967231109, -120.21159890801503],
             [-120.21158734028334, -120.2115781045798,
              -120.21158043660672, -120.21158967231109],
             [-120.2115781045798, -120.2115688688758,
              -120.21157120090189, -120.21158043660672],
             [-120.2115688688758, -120.21155963317135,
              -120.21156196519657, -120.21157120090189]],
            [[-120.21161737942151, -120.21160814371851,
              -120.21161047574842, -120.21161971145229],
             [-120.21160814371851, -120.21159890801503,
              -120.21160124004409, -120.21161047574842],
             [-120.21159890801503, -120.21158967231109,
              -120.21159200433934, -120.21160124004409],
             [-120.21158967231109, -120.21158043660672,
              -120.21158276863409, -120.21159200433934],
             [-120.21158043660672, -120.21157120090189,
              -120.21157353292838, -120.21158276863409],
             [-120.21157120090189, -120.21156196519657,
              -120.21156429722222, -120.21157353292838]],
            [[-120.21161971145229, -120.21161047574842,
              -120.21161280777879, -120.2116220434835],
             [-120.21161047574842, -120.21160124004409,
              -120.21160357207363, -120.21161280777879],
             [-120.21160124004409, -120.21159200433934,
              -120.21159433636801, -120.21160357207363],
             [-120.21159200433934, -120.21158276863409,
              -120.21158510066192, -120.21159433636801],
             [-120.21158276863409, -120.21157353292838,
              -120.21157586495535, -120.21158510066192],
             [-120.21157353292838, -120.21156429722222,
              -120.21156662924835, -120.21157586495535]]]
        latb[:] = [
            [[22.748507533242535, 22.748509683865187,
              22.748518201130356, 22.74851605050742],
             [22.748509683865187, 22.74851183448702,
              22.748520351752475, 22.748518201130356],
             [22.74851183448702, 22.74851398510802,
              22.74852250237377, 22.748520351752475],
             [22.74851398510802, 22.748516135728206,
              22.74852465299425, 22.74852250237377],
             [22.748516135728206, 22.748518286347593,
              22.748526803613903, 22.74852465299425],
             [22.748518286347593, 22.748520436966125,
              22.748528954232754, 22.748526803613903]],
            [[22.74851605050742, 22.748518201130356,
              22.748526718395773, 22.74852456777256],
             [22.748518201130356, 22.748520351752475,
              22.748528869018187, 22.748526718395773],
             [22.748520351752475, 22.74852250237377,
              22.748531019639763, 22.748528869018187],
             [22.74852250237377, 22.74852465299425,
              22.748533170260536, 22.748531019639763],
             [22.74852465299425, 22.748526803613903,
              22.74853532088048, 22.748533170260536],
             [22.748526803613903, 22.748528954232754,
              22.748537471499613, 22.74853532088048]],
            [[22.74852456777256, 22.748526718395773,
              22.74853523566145, 22.748533085037966],
             [22.748526718395773, 22.748528869018187,
              22.74853738628417, 22.74853523566145],
             [22.748528869018187, 22.748531019639763,
              22.748539536906023, 22.74853738628417],
             [22.748531019639763, 22.748533170260536,
              22.7485416875271, 22.748539536906023],
             [22.748533170260536, 22.74853532088048,
              22.748543838147327, 22.7485416875271],
             [22.74853532088048, 22.748537471499613,
              22.748545988766764, 22.748543838147327]],
            [[22.748533085037966, 22.74853523566145,
              22.74854375292739, 22.74854160230359],
             [22.74853523566145, 22.74853738628417,
              22.748545903550376, 22.74854375292739],
             [22.74853738628417, 22.748539536906023,
              22.748548054172538, 22.748545903550376],
             [22.748539536906023, 22.7485416875271,
              22.748550204793883, 22.748548054172538],
             [22.7485416875271, 22.748543838147327,
              22.7485523554144, 22.748550204793883],
             [22.748543838147327, 22.748545988766764,
              22.748554506034104, 22.7485523554144]],
            [[22.74854160230359, 22.74854375292739,
              22.74855227019359, 22.748550119569494],
             [22.74854375292739, 22.748545903550376,
              22.748554420816852, 22.74855227019359],
             [22.748545903550376, 22.748548054172538,
              22.74855657143929, 22.748554420816852],
             [22.748548054172538, 22.748550204793883,
              22.74855872206093, 22.74855657143929],
             [22.748550204793883, 22.7485523554144,
              22.748560872681754, 22.74855872206093],
             [22.7485523554144, 22.748554506034104,
              22.748563023301763, 22.748560872681754]]]
        return tncf

    def testVal2idx(self):
        ncf = PseudoNetCDFFile()
        coorde = np.arange(9, dtype='f')
        coordc = (coorde[:-1] + coorde[1:]) / 2.
        ncf.createDimension('coord', coordc.size)
        ncf.createDimension('nv', 2)
        ncf.createVariable(
            'coord', 'f', ('coord',), values=coordc
        )
        bncf = ncf.copy()
        bncf.createVariable(
            'coord_bounds', 'f', ('coord', 'nv'),
            values=coorde.repeat(2, 0)[1:-1].reshape(-1, 2)
        )
        bncf.variables['coord'].bounds = 'coord_bounds'
        cvals = [-1, .25, 4.5, 4.99, 5, 6.75, 10]
        expectedb = np.array([0, 0, 4, 4, 5, 6, 7])
        expectedn = np.array([0, 0, 4, 4, 4, 6, 7])

        def checkvals(ncf, method, clean, compare):
            withbnds = str('coord_bounds' in ncf.variables)
            prefix = withbnds + '&' + method + '&' + clean
            idx = ncf.val2idx(
                'coord', cvals,
                method=method, clean=clean, bounds='warn'
            )
            warn(prefix + ' got: ' + repr(idx))
            warn(prefix + ' chk: ' + repr(compare))
            assert(np.ma.allclose(compare, idx))

        mw = np.ma.masked_where

        nn_mask = [0] * 7
        bn_mask = [0] * 7
        nm_mask = [1, 1, 0, 0, 0, 0, 1]
        bm_mask = [1, 0, 0, 0, 0, 0, 1]
        em_mask = [1, 1, 0, 1, 1, 1, 1]
        checkvals(ncf, 'nearest', 'none', mw(nn_mask, expectedn))
        checkvals(ncf, 'bounds', 'none', mw(bn_mask, expectedb))
        checkvals(ncf, 'nearest', 'mask', mw(nm_mask, expectedn))
        checkvals(ncf, 'bounds', 'mask', mw(bm_mask, expectedb))

        checkvals(bncf, 'nearest', 'none', mw(nn_mask, expectedn))
        checkvals(bncf, 'bounds', 'none', mw(bn_mask, expectedb))
        checkvals(bncf, 'nearest', 'mask', mw(nm_mask, expectedn))
        checkvals(bncf, 'bounds', 'mask', mw(bm_mask, expectedb))

        checkvals(bncf, 'exact', 'mask', mw(em_mask, expectedb))

    def testCopyVariable(self):
        tncf = self.testncf
        var = tncf.copyVariable(
            tncf.variables['O3'], key='O3_PPB', withdata=True)
        self.assertEqual(True, (var[:] == tncf.variables['O3_PPB']).all())

    def testSubsetVariables(self):
        tncf = self.testncf.copy()
        var = tncf.copyVariable(
            tncf.variables['O3'], key='O3_PPB', withdata=True)
        var[:] *= 1e3
        var = tncf.copyVariable(
            tncf.variables['O3'], key='O3_PPT', withdata=True)
        var[:] *= 1e6
        sncf = tncf.subsetVariables(['O3_PPT'])
        self.assertEqual(len(sncf.variables), 1)
        self.assertEqual(set(sncf.variables), set(['O3_PPT']))
        sncf = tncf.subsetVariables(['O3_PPT'], exclude=True)
        self.assertEqual(len(sncf.variables), len(
            self.myvars.union(['O3_PPB'])))
        self.assertEqual(set(sncf.variables), self.myvars.union(['O3_PPB']))

    def testRenameVariables(self):
        tncf = self.testncf
        sncf = tncf.renameVariables(O3='O3_PPM')
        self.assertEqual(len(sncf.variables), len(self.myvars))
        self.assertEqual(set(sncf.variables), self.mymeta.union(['O3_PPM']))

    def testRenameDimensions(self):
        tncf = self.testncf
        sncf = tncf.renameDimensions(TSTEP='TIME')
        self.assertEqual(len(sncf.dimensions), len(tncf.dimensions))
        self.assertEqual(set(sncf.dimensions), set(self.mydims))

    def testSliceDimensionInt(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        sncf = tncf.sliceDimensions(TSTEP=0)
        self.assertEqual(len(sncf.dimensions['TSTEP']), 1)
        self.assertEqual(
            True, (sncf.variables['O3'][:] == o3[0]).all())

    def testSliceDimensionList(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        sncf = tncf.sliceDimensions(TSTEP=[0])
        self.assertEqual(len(sncf.dimensions['TSTEP']), 1)
        self.assertEqual(
            True, (sncf.variables['O3'][:] == o3[0]).all())
        sncf = tncf.sliceDimensions(TSTEP=[0, 8])
        self.assertEqual(len(sncf.dimensions['TSTEP']), 2)
        self.assertEqual(
            True, (sncf.variables['O3'][:] == o3[[0, 8]]).all())

    def testSliceDimensionComboListInt(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        sncf = tncf.sliceDimensions(TSTEP=[0, 8], ROW=2, COL=3)
        self.assertEqual(len(sncf.dimensions['TSTEP']), 2)
        self.assertEqual(len(sncf.dimensions['ROW']), 1)
        self.assertEqual(len(sncf.dimensions['COL']), 1)
        self.assertEqual(True, (sncf.variables['O3'][:] == o3[[
                         0, 8], :, 2, 3][:, :, None, None]).all())

    def testSliceDimensionMultiArray(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        i = np.arange(4)
        sncf = tncf.sliceDimensions(TSTEP=i, LAY=i, ROW=i, COL=i)
        self.assertEqual(len(sncf.dimensions['POINTS']), 4)
        self.assertEqual(
            True, (sncf.variables['O3'][:] == o3[i, i, i, i]).all())

    def testSliceDimensionSlice(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        sncf = tncf.sliceDimensions(ROW=slice(0, 3))
        self.assertEqual(len(sncf.dimensions['ROW']), 3)
        self.assertEqual(
            True, (sncf.variables['O3'][:] == o3[:, :, 0:3, :]).all())

    def testSliceDimensionMultiSlice(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        sncf = tncf.sliceDimensions(
            LAY=slice(0, 3), ROW=slice(0, 3), COL=slice(0, 3))
        self.assertEqual(len(sncf.dimensions['LAY']), 3)
        self.assertEqual(len(sncf.dimensions['ROW']), 3)
        self.assertEqual(len(sncf.dimensions['COL']), 3)
        self.assertEqual(
            True, (sncf.variables['O3'][:] == o3[:, 0:3, 0:3, 0:3]).all())

    def testApplyAlongDimensionsNamed(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        ancf = tncf.applyAlongDimensions(LAY='min')
        self.assertEqual(
            True, (ancf.variables['O3'][:] == o3.min(1, keepdims=True)).all())

    def testApplyAlongDimensionsMultiNamed(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        ancf = tncf.applyAlongDimensions(LAY='min', ROW='max')
        self.assertEqual(True, (ancf.variables['O3'][:] == o3.min(
            1, keepdims=True).max(2, keepdims=True)).all())

    def testApplyAlongDimensionsConvolve(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        ancf = tncf.applyAlongDimensions(TSTEP=lambda x: np.convolve(
            x, np.ones(2, dtype='f') / 2., mode='valid'))
        co3 = (o3[1:] + o3[:-1]) / 2
        self.assertEqual(True, (ancf.variables['O3'][:] == co3).all())

    def testApplyAlongDimensionsConvolveWithMax(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:]
        co3 = (o3[1:] + o3[:-1]) / 2
        # Testing convolution; useful for mda8

        def convf(x):
            return np.convolve(x, np.ones(2, dtype='f') / 2., mode='valid')

        ancf = tncf.applyAlongDimensions(TSTEP=convf)\
                   .applyAlongDimensions(TSTEP=np.max)
        mco3 = co3.max(0, keepdims=True)
        self.assertEqual(True, (ancf.variables['O3'][:] == mco3).all())

    def testAdd(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:].copy()
        tmpf = tncf + tncf
        np_all_close(tmpf.variables['O3'][:], 2 * o3)

    def testSub(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:].copy()
        tmpf = tncf - tncf
        np_all_close(tmpf.variables['O3'][:], o3 - o3)

    def testDiv(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:].copy()
        np.seterr(invalid='ignore')
        tmpf = tncf / tncf
        np_all_close(tmpf.variables['O3'][:], o3 / o3)
        np.seterr(invalid='warn')

    def testMul(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:].copy()
        tmpf = tncf * tncf
        np_all_close(tmpf.variables['O3'][:], o3**2)

    def testPow(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:].copy()
        np.seterr(over='ignore', invalid='ignore')
        tmpf = tncf ** tncf
        np_all_close(tmpf.variables['O3'][:], o3**o3)
        np.seterr(over='warn', invalid='warn')

    def testMod(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:].copy()
        np.seterr(divide='ignore')
        tmpf = tncf % tncf
        np_all_close(tmpf.variables['O3'][:], o3 % o3)
        np.seterr(divide='warn')

    def testXAdd(self):
        tncf = self.testncf
        o3 = tncf.variables['O3'][:].copy()
        tmpf = tncf.copy()
        print(tmpf._operator_exclude_vars)
        tmpf.setCoords(['O3'])
        print(tmpf._operator_exclude_vars)
        tmpf = tmpf + tncf
        np_all_close(tmpf.variables['O3'][:], o3)

    @requires_basemap
    def testGetMap(self):
        tncf = self.testncf
        tncf.getMap(maptype='basemap_auto')

    @requires_pyproj
    def testGetproj(self):
        tncf = self.testncf
        tncf.getproj(withgrid=False, projformat='pyproj')

    @requires_pyproj
    def testLl2xy(self):
        tncf = self.testncf
        crs = tncf.variables['lambert_conformal_conic']
        y0 = crs.false_northing
        x0 = crs.false_easting
        lon0, lat0 = tncf.xy2ll(x0, y0)
        x0t, y0t = tncf.ll2xy(lon0, lat0)
        self.assertEqual((x0, y0), (x0t, y0t))

    @requires_pyproj
    def testLl2ij(self):
        tncf = self.testncf
        y0 = 0
        x0 = 0
        lon0, lat0 = tncf.xy2ll(x0, y0)
        i0, j0 = tncf.ll2ij(lon0, lat0)
        self.assertEqual(0, i0)
        self.assertEqual(0, j0)

    def testTime2t(self):
        from datetime import datetime
        time = np.array([
            datetime.strptime('1970-01-{}+0000'.format(t), '%Y-%m-%d %H:%M%z')
            for t in [
                '01 00:20', '01 01:20', '01 01:40', '01 06:00', '02 01:00'
            ]
        ])
        tncf = self.testncf
        t = tncf.time2t(time[:])
        np_all_close(t, [0, 1, 2, 6, 23])
        t = tncf.time2t(time[:-1], ttype='bounds')
        np_all_close(t, [0, 1, 1, 6])
        t = tncf.time2t(time[:], ttype='bounds_close')
        self.assertEqual(True, np.allclose(t, [0, 1, 1, 6, 23]))
        t = tncf.time2t(time[:], ttype='bounds')
        np_all_close(t, np.ma.masked_invalid([0, 1, 1, 6, 23]))
        self.assertEqual(t.mask[-1], True)

    def testTime2idx(self):
        from datetime import datetime
        time = np.array([
            datetime.strptime('1970-01-{}+0000'.format(t), '%Y-%m-%d %H:%M%z')
            for t in [
                '01 00:20', '01 01:20', '01 01:40', '01 06:00', '02 01:00'
            ]
        ])
        tncf = self.testncf
        t = tncf.time2idx(time[:])
        np_all_close(t, [0, 1, 2, 6, 23])
        t = tncf.time2idx(
            time, method='bounds', clean='none',
            left=-1, right=999,
        )
        np_all_close(t, [0, 1, 1, 6, 999])
        t = tncf.time2idx(time[:], method='bounds')
        self.assertEqual(True, np.allclose(t, [0, 1, 1, 6, 23]))
        t = tncf.time2idx(time[:], method='bounds', clean='mask', right=np.nan)
        np_all_close(t, np.ma.masked_invalid([0, 1, 1, 6, np.nan]))
        self.assertEqual(t.mask[-1], True)

    @requires_pyproj
    def testXy2ll(self):
        tncf = self.testncf
        crs = tncf.variables['lambert_conformal_conic']
        y0 = crs.false_northing
        x0 = crs.false_easting
        lonmid, latmid = tncf.xy2ll(x0, y0)
        self.assertEqual(True, np.allclose(
            lonmid, crs.longitude_of_central_meridian))
        self.assertEqual(True, np.allclose(
            latmid, crs.latitude_of_projection_origin))
        lon0, lat0 = tncf.xy2ll(0, 0)
        self.assertEqual(True, np.allclose(
            lon0, tncf.variables['longitude_bounds'][0, 0, 0]))
        self.assertEqual(True, np.allclose(
            lat0, tncf.variables['latitude_bounds'][0, 0, 0]))

    @requires_pyproj
    def testIj2ll(self):
        tncf = self.testncf
        lon0, lat0 = tncf.ij2ll(0, 0)
        self.assertEqual(True, np.allclose(
            lon0, tncf.variables['longitude'][0, 0]))
        self.assertEqual(True, np.allclose(
            lat0, tncf.variables['latitude'][0, 0]))

    def testEval(self):
        tncf = self.testncf.copy()
        tncf.eval('O3_PPB = O3 * 1000.', inplace=True, copyall=False)
        o3ppmv = tncf.variables['O3']
        o3ppbv = tncf.variables['O3_PPB']
        self.assertEqual(True, ((o3ppmv * 1000.) == o3ppbv).all())

    def testSetncatts(self):
        tncf = self.testncf.copy()
        tncf.setncatts({'test_new1': 1, 'test_new2': 'five'})
        self.assertEqual(tncf.test_new1, 1)
        self.assertEqual(tncf.test_new2, 'five')

    def testGetncatts(self):
        tncf = self.testncf
        t = tncf.getncatts()
        self.assertEqual(t['str_one'], '1')
        self.assertEqual(t['int_two'], 2)
        self.assertEqual(t['float_threeptfive'], 3.5)

    def testCopy(self):
        tncf = self.testncf
        nncf = tncf.copy(props=True, dimensions=True,
                         variables=True, data=True)
        for pk in tncf.ncattrs():
            self.assertEqual(getattr(tncf, pk), getattr(nncf, pk, None))

        for dk, dv in tncf.dimensions.items():
            dvl = len(dv)
            ndvl = len(nncf.dimensions[dk])
            self.assertEqual(dvl, ndvl)

        for vk, vv in tncf.variables.items():
            nvv = nncf.variables[vk]
            self.assertEqual(True, np.allclose(vv[...], nvv[...]))
            self.assertEqual(vv.dimensions, nvv.dimensions)
            self.assertEqual(vv.dtype.char, nvv.dtype.char)
            for pk in vv.ncattrs():
                pv = getattr(vv, pk)
                npv = getattr(nvv, pk, None)
                testv = pv == npv
                self.assertEqual(True, np.any(testv))

    def testGetTimes(self):
        tncf = self.testncf
        t = tncf.getTimes()
        self.assertEqual(True, (t == self.mytimes).all())

    def testStack(self):
        tncf = self.testncf
        sncf = tncf.stack(tncf, 'TSTEP')
        to3 = tncf.variables['O3'][:]
        no3 = sncf.variables['O3'][:]
        origlen = len(tncf.dimensions['TSTEP'])
        self.assertEqual(origlen * 2, len(sncf.dimensions['TSTEP']))
        self.assertEqual(True, np.allclose(to3, no3[:origlen]))
        self.assertEqual(True, np.allclose(to3, no3[origlen:]))

    def testRemoveSingleton(self):
        tncf = self.testncf
        nncf = tncf.copy()
        nncf.createDimension('test', 1)
        nncf = nncf.removeSingleton(dimkey='test')
        self.assertEqual(set(tncf.dimensions), set(nncf.dimensions))
        nncf.createDimension('test', 1)
        nncf = nncf.removeSingleton(dimkey=None)
        self.assertEqual(set(tncf.dimensions), set(nncf.dimensions))

    def testCreateDimension(self):
        tncf = self.testncf.copy()
        ndims = len(tncf.dimensions)
        olddims = set(tncf.dimensions)
        newdims = olddims.union(['newd'])
        tncf.createDimension('newd', 5)
        self.assertEqual(len(tncf.dimensions), ndims + 1)
        self.assertEqual(set(tncf.dimensions), newdims)

    def testCreateVariable(self):
        tncf = self.testncf.copy()
        nvars = len(tncf.variables)
        var = tncf.createVariable('test', 'f', ('TSTEP',))
        self.assertEqual(len(tncf.variables), nvars + 1)
        self.assertEqual(var.dimensions, ('TSTEP',))
        self.assertEqual(var.dtype.char, 'f')

    def testSave(self):
        pass

    def testNcattrs(self):
        tncf = self.testncf
        tncf.ncattrs()

    def testSetncattr(self):
        tncf = self.testncf
        tncf.setncattr('test', 1)
        self.assertEqual(tncf.test, 1)

    def testDelncattr(self):
        tncf = self.testncf.copy()
        tncf.setncattr('test', 1)
        tncf.delncattr('test')
        self.assertEqual(False, hasattr(tncf, 'test'))

    def testInsertDimension(self):
        tncf = self.testncf.copy()
        nncf = tncf.insertDimension(TEST=2, before='LAY')
        checkv = (nncf.variables['O3'][:, 0] == nncf.variables['O3'][:, 1])
        self.assertEqual(True, checkv.all())
        nncf = tncf.insertDimension(TEST=2)
        checkv = (nncf.variables['O3'][0, :] == nncf.variables['O3'][1, :])
        self.assertEqual(True, checkv.all())

    def testInterpDimension(self):
        f1 = PseudoNetCDFFile()
        f1.createDimension('time', 2)
        f1.createDimension('layer', 3)
        f1.createDimension('latitude', 4)
        f1.createDimension('longitude', 5)
        lay = f1.createVariable('layer', 'f', ('layer',))
        lay[:] = np.arange(0, 3)
        simple = f1.createVariable(
            'simple', 'f', ('time', 'layer', 'latitude', 'longitude'))
        simple[0] = np.arange(3 * 4 * 5).reshape(3, 4, 5)
        simple[1] = np.arange(3 * 4 * 5).reshape(3, 4, 5)

        f2 = f1.applyAlongDimensions(layer=lambda x: (x[1:] + x[:-1]) * .5)
        f3 = f1.interpDimension('layer', f2.variables['layer'])
        self.assertEqual(True, np.allclose(
            f2.variables['simple'][:], f3.variables['simple'][:]))
        f4 = PseudoNetCDFFile()
        f4.createDimension('time', 2)
        f4.createDimension('layer', 3)
        f4.createDimension('latitude', 4)
        f4.createDimension('longitude', 5)
        lay = f4.createVariable(
            'layer', 'f', ('time', 'layer', 'latitude', 'longitude'))
        lay[:] = np.arange(0, 3)[None, :, None, None]
        simple = f4.createVariable(
            'simple', 'f', ('time', 'layer', 'latitude', 'longitude'))
        simple[0] = np.arange(3 * 4 * 5).reshape(3, 4, 5)
        simple[1] = np.arange(3 * 4 * 5).reshape(3, 4, 5)

        f5 = f4.applyAlongDimensions(layer=lambda x: (x[1:] + x[:-1]) * .5)
        lay[1] += .25
        f6 = f4.interpDimension('layer', f5.variables['layer'])
        self.assertEqual(True, np.allclose(
            f5.variables['simple'][0], f6.variables['simple'][0]))
        self.assertEqual(False, np.allclose(
            f5.variables['simple'][1], f6.variables['simple'][1]))

    def testNetCDFFileNew(self):
        t = PseudoNetCDFFile.__new__(PseudoNetCDFFile)
        self.assertEqual(t.variables, {})
        self.assertEqual(t.dimensions, {})
        self.assertEqual(t.ncattrs(), ())

    def testNetCDFFileInit(self):
        from numpy import arange
        self._makencf()
        tncf = self.testncf
        self.assertEqual(len(tncf.dimensions['TSTEP']), 24)
        self.assertEqual(len(tncf.dimensions['LAY']), 4)
        self.assertEqual(len(tncf.dimensions['ROW']), 5)
        self.assertEqual(len(tncf.dimensions['COL']), 6)
        self.assertEqual(len(tncf.dimensions['nv']), 4)

        tncf.fish = 2
        setattr(tncf, 'FROG-DOG', 'HAPPY')

        self.assertEqual(set(tncf.variables.keys()), self.myvars)
        o3 = tncf.variables['O3']
        self.assertEqual(
            True, (o3 == arange(24 * 4 * 5 * 6).reshape(24, 4, 5, 6)).all())

        self.assertEqual(o3.typecode(), 'f')

        filedims = list(tncf.dimensions)
        filedims.sort()
        vardims = list(o3.dimensions)
        vardims.sort()
        filedims.remove('nv')
        filedims.remove('tnv')

        self.assertEqual(filedims, vardims)
        from PseudoNetCDF.pncgen import Pseudo2NetCDF
        n = Pseudo2NetCDF().convert(tncf)
        self.assertEqual(set(n.variables.keys()), self.myvars)
        self.assertEqual(
            dict([(k, len(v)) for k, v in n.dimensions.items()]),
            dict([(k, len(v)) for k, v in self.testncf.dimensions.items()])
        )
        self.assertEqual(
            True, (n.variables['O3'][...] == tncf.variables['O3'][...]).all())
        self.assertEqual(n.variables['O3'].units, 'ppbv')
        self.assertEqual(n.fish, 2)
        self.assertEqual(getattr(n, 'FROG-DOG'), 'HAPPY')

    def testNetCDFVariables(self):
        from numpy import arange
        tncf = PseudoNetCDFFile()
        tncf.createDimension('TSTEP', 24)
        tncf.createDimension('LAY', 4)
        tncf.createDimension('ROW', 5)
        tncf.createDimension('COL', 6)

        def const(*args, **kwds):
            vals = arange(24 * 4 * 5 * 6).reshape((24, 4, 5, 6))
            dims = ('TSTEP', 'LAY', 'ROW', 'COL')
            return PseudoNetCDFVariable(tncf, args[0], 'f', dims, values=vals)
        tncf.variables = PseudoNetCDFVariables(const, ['NO', 'O3'])
        self.assertEqual(True, (tncf.variables['O3'] == arange(
            24 * 4 * 5 * 6).reshape(24, 4, 5, 6)).all())

    def runTest(self):
        pass
